local MAJOR = "StarLibPluginTalents-1.0"
local MINOR = 13
local PluginTalents = LibStub:NewLibrary(MAJOR, MINOR)
if not PluginTalents then return end
local TalentQuery = LibStub("LibTalentQuery-1.0", true)
assert(TalentQuery, MAJOR .. " requires LibTalentQuery-1.0")
local LibTimer = LibStub("StarLibTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")

local _G = _G
local GameTooltip = _G.GameTooltip
local UnitIsUnit = _G.UnitIsUnit
local GetNumTalentTabs = _G.GetNumTalentTabs
local GetTalentTabInfo = _G.GetTalentTabInfo
local UnitExists = _G.UnitExists
local UnitIsPlayer = _G.UnitIsPlayer
local UnitName = _G.UnitName
local expired
local expireTimer
local EXPIRE_TIME = 5000
local spec = setmetatable({}, {__mode='v'})
local frame = CreateFrame("Frame")
local count = 0
local query = {}

local L = setmetatable({}, {__index = function(self, key)
	str = rawget(self, key)
	if str then
		return str
	else
		return key
	end
end})

if not PluginTalents.__index then
	PluginTalents.__index = PluginTalents
end

-- @name PluginTalents:New
-- @usage :New(environment)
-- @parma environment This will be the environment when loadstring is callled.
-- @return A new PluginTalents object
function PluginTalents:New(environment)

	environment.SpecText = self.SpecText
	environment.GetSpec = self.GetSpec
	environment.ClearSpec = self.ClearSpec
	
	TalentQuery.RegisterCallback(self, "TalentQuery_Ready")
end

local talentTrees = {
	[L["Druid"]] = {L["Balance"], L["Feral Combat"], L["Restoration"]},
	[L["Hunter"]] = {L["Beast Mastery"], L["Marksmanship"], L["Survival"]},
	[L["Mage"]] = {L["Arcane"], L["Fire"], L["Frost"]},
	[L["Paladin"]] = {L["Holy"], L["Protection"], L["Retribution"]},
	[L["Priest"]] = {L["Discipline"], L["Holy"], L["Shadow"]},
	[L["Rogue"]] = {L["Assassination"], L["Combat"], L["Subtlety"]},
	[L["Shaman"]] = {L["Elemental"], L["Enhancement"], L["Restoration"]},
	[L["Warlock"]] = {L["Affliction"], L["Demonology"], L["Destruction"]},
	[L["Warrior"]] = {L["Arms"], L["Fury"], L["Protection"]},
}

local indexOf = function(t, val)
	for i=1, #t do
		if t[i] == val then
			return i
		end
	end
end

local indicesOf = function(t, val)
	local a = {}
	for i=1, #t do
		if t[i] == val then
			tinsert(a, i)
		end
	end
	return unpack(a)
end

local function sortfunc(a,b) return a>b end

function PluginTalents:TalentQuery_Ready(e, name, realm, unitid)
	if not TalentQuery or not unitid then return end
	local guid = UnitGUID(unitid)
	local isnotplayer = not UnitIsUnit("player", unitid)
	if not spec[guid] then
		spec[guid] = {[4]=NONE}
		local highPoints = {}
		local specNames = {}
		local group = GetActiveTalentGroup(isnotplayer)
		for tab = 1, GetNumTalentTabs(isnotplayer) do
			local treename, _, pointsspent = GetTalentTabInfo(tab, isnotplayer, nil, group)
			highPoints[tab] = pointsspent
			spec[guid][tab] = pointsspent
			specNames[tab] = treename
		end
		if highPoints[1] == nil or highPoints[2] == nil or highPoints[3] == nil then spec[guid] = nil return end
		table.sort(highPoints, sortfunc)
		local first, second = select(1, indicesOf(spec[guid], highPoints[1])), select(2, indicesOf(spec[guid], highPoints[1]))
		if highPoints[1] > 0 and highPoints[2] > 0 and highPoints[1] - highPoints[2] <= 5 and highPoints[1] ~= highPoints[2] then
			spec[guid][4] = specNames[indexOf(spec[guid], highPoints[1])] .. "/" .. specNames[indexOf(spec[guid], highPoints[2])]
		elseif highPoints[1] > 0 and first and second then
			spec[guid][4] = specNames[first] .. "/" .. specNames[second]
		elseif highPoints[1] > 0 then
			spec[guid][4] = specNames[indexOf(spec[guid], highPoints[1])]
		end
	end
	query[guid] = false
	if expireTimer then expireTimer:Stop() end
	if expireExpireTimer then expireExpireTimer:Stop() end
	expired = false
end

local function expireExpire()
	expired = false
end

local expireExpireTimer
local function expireQuery()
	expireExpireTimer = expireExpireTimer or LibTimer:New(MAJOR .. " expireExpireTimer", 2000, false, expireExpire)
	expireExpireTimer:Start()
	expired = true
	wipe(query)
end

function PluginTalents.SendQuery(unit)
	if not TalentQuery or not UnitIsPlayer(unit) then return end

	if UnitIsUnit(unit, "player") then
		PluginTalents.TalentQuery_Ready(_, UnitName(unit))
	else
		--TalentQuery:NotifyInspect("mouseover")
		TalentQuery:Query(unit)
		if expireTimer then
			expireTimer:Stop()
		end
		expireTimer = expireTimer or LibTimer:New(MAJOR .. " expireTimer", EXPIRE_TIME, false, expireQuery)
		expireTimer:Start()
	end
	count = 0
	local guid = UnitGUID(unit)
	query[guid] = true
end

local function newUnit()
	if expireTimer then expireTimer:Stop() end
	if expireExpireTimer then expireExpireTimer:Stop() end
	expired = false
end

GameTooltip:HookScript("OnTooltipSetUnit", newUnit)

function PluginTalents.SpecText(unit)
	if type(unit) ~= "string" then return end
	if not UnitExists(unit) then return L["Unknown"] end
	if not UnitIsPlayer(unit) then return end
	local guid = UnitGUID(unit)
	
	if expired then return L["Expired"] end
	
	if not query[guid] then 
		if expireTimer then expireTimer:Stop() end
		if expireExpireTimer then expireExpireTimer:Stop() end
		PluginTalents.SendQuery(unit)
	end
	
	local periods = ""
	for i = 0, count % 3 do
		periods = periods .. "."
	end
	count = count + 1
	if not spec[guid] then return L["Loading"] .. periods end
	return ('%s (%d/%d/%d)'):format(spec[guid][4], spec[guid][1], spec[guid][2], spec[guid][3])
end

function PluginTalents.GetSpec(unit)
	if type(unit) ~= "string" or not UnitExists(unit) then return end
	local guid = UnitGUID(unit)
	
	if not query[guid] then
		if expireTimer then expireTimer:Stop() end
		if expireExpireTimer then expireExpireTimer:Stop() end
		PluginTalents.SendQuery(unit)
	end
	
	if not spec[guid] then return nil end
	return unpack(spec[guid])
end

function PluginTalents.ClearSpec(unit)
	if type(unit) ~= "string" or not UnitExists(unit) then return end
	local guid = UnitGUID(unit)
	spec[guid] = nil
end


