local MAJOR = "StarLibWidget-1.0" 
local MINOR = 13

assert(LibStub, MAJOR.." requires LibStub") 
local LibWidget = LibStub:NewLibrary(MAJOR, MINOR)
if not LibWidget then return end
local LibProperty = LibStub("StarLibProperty-1.0", true)
assert(LibProperty, MAJOR .. " requires StarLibProperty-1.0")
local LibTimer = LibStub("StarLibTimer-1.0", true)
assert(LibTimer, MAJOR .. " requires StarLibTimer-1.0")
local LibError = LibStub("StarLibError-1.0", true)
assert(LibError, MAJOR .. " requires StarLibError-1.0")
local utils = LibStub("StarLibUtils-1.0", true)
assert(utils, MAJOR .. " requires StarLibUtils-1.0")

local pool = setmetatable({}, {__mode = "k"})

if not LibWidget.__index then
	LibWidget.__index = LibWidget
end

local function rfind(str, char)
	local i = strlen(str)
	while i > 0 do
		if str:sub(i, i) == char then
			return i
		end
		i = i - 1
	end
end

-- Create a new StarLibWidget object
-- @name WidgetText.New
-- @usage WidgetText:New(visitor, name, config, row, col, layer, type, environment, errorLevel)
-- @param visitor An StarLibCore-1.0 object
-- @parma name This widget's name
-- @param config This widget's configuration
-- @param row This widget's row
-- @param col This widget's column
-- @param layer This widget's layer
-- @param type Dict of widget types.
-- @param errorLevel The error level for this object.
-- @return A new StarLibWidgetText object
function LibWidget:New(child, visitor, name, config, row, col, layer, typeOf, errorLevel) 
	
	assert(type(child) == "table", "No child")
	assert(type(visitor) == "table", "No visitor")
	assert(type(name) == "string", "No name")
	assert(type(config) == "table", "No config")
	assert(type(row) == "number", "No row")
	assert(type(col) == "number", "No col")
	assert(type(typeOf) == "table", "No type")
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end
	
	obj.child = child
	obj.visitor = visitor
	obj.name = name
	obj.config = config
	obj.persistent = config.persistent
	obj.row = row
	obj.col = col
	obj.layer = layer
	obj.type = typeOf
	obj.errorLevel = errorLevel
	obj.lcd = visitor.lcd
	
	local pos1 = name:find(":")
	local pos2 = rfind(name, ":")
	if pos1 then
		obj.layout_base = name:sub(0, pos1 - 1)
		obj.widget_base = name:sub(pos1 + 1, pos2 - 1)	
	end
	obj.started = false
	obj.errorLevel = errorLevel or 3
	obj.error = LibError:New(MAJOR .. ": " .. name, errorLevel)

	for k, v in pairs(obj) do
		child[k] = v
	end
	
	setmetatable(obj, self)

	obj.deleted = false
	
	return obj	
end


-- @name StarLibWidget:Del
-- @usage StarLibWidget:Del([widget]) or object:Del()
-- @param widget An optional widget object
-- @return Nothing
function LibWidget:Del(widget)
	if not widget then widget = self end
	assert(not widget.deleted)
	pool[widget] = true
	widget.deleted = true
	widget.error:Del()
end
    
local anchors = {
	"TOP",
	"TOPRIGHT",
	"TOPLEFT",
	"BOTTOM",
	"BOTTOMRIGHT",
	"BOTTOMLEFT",
	"RIGHT",
	"LEFT",
	"CENTER"
}

local anchorsDict = {}

for i, v in ipairs(anchors) do
	anchorsDict[v] = i
end
	
local strataNameList = {
	"TOOLTIP", "FULLSCREEN_DIALOG", "FULLSCREEN", "DIALOG", "HIGH", "MEDIUM", "LOW", "BACKGROUND"
}

local strataLocaleList = {
	"Tooltip", "Fullscreen Dialog", "Fullscreen", "Dialog", "High", "Medium", "Low", "Background"
}
	
function LibWidget:GetOptions(db, callback, data)
	local options = {
		name = "Frame Details",
		type = "group",
		args = {
			add = {
				name = "Add Point",
				desc = "Add a new point",
				type = "input",
				set = function(info, v)
					tinsert(db.points, {"Center", "GameTooltip", "Center", 0, 0})
					if type(callback) == "function" then
						callback(data)
					end
				end,
				order = 1
			},
			parent = {
				name = "Frame Parent",
				type = "input",
				get = function() return db.parent or "GameTooltip" end,
				set = function(info, val) 
					db.parent = val; 
					db["parentDirty"] = true 
					if type(callback) == "function" then
						callback(data)
					end
				end,
				order = 5
			},
			strata = {
				name = "Frame Strata",
				type = "select",
				values = strataLocaleList,
				get = function()
					return db.strata or 1
				end,
				set = function(info, val)
					db.strata = val
					db["strataDirty"] = true
					if type(callback) == "function" then
						callback(data)
					end
				end,
				order = 6
			},
			level = {
				name = "Frame Level",
				type = "input",
				pattern = "%d",
				get = function()
					return tostring(db.level or 1)
				end,
				set = function(info, val)
					db.level = tonumber(val)
					db["levelDirty"] = true
					if type(callback) == "function" then
						callback(data)
					end
				end,
				order = 7
			}
		}
	}
	for i, point in ipairs(db.points) do
			options.args["point" .. i] = {
				name = "Point #" .. i,
				type = "group",
				args = {
					point = {
						name = "Text anchor",
						type = "select",
						values = anchors,
						get = function() return anchorsDict[point[1] or 1] end,
						set = function(info, v) 
							point[1] = anchors[v]; 
							db["pointsDirty"] = true 
							if type(callback) == "function" then
								callback(data)
							end							
						end,
						order = 1
					},
					relativeFrame = {
						name = "Relative Frame",
						type = "input",
						get = function() return point[2] end,
						set = function(info, v) 
							point[2] = v; 
							db["pointsDirty"] = true 
							if type(callback) == "function" then
								callback(data)
							end							
						end,
						order = 2
					},
					relativePoint = {
						name = "Relative Point",
						type = "select",
						values = anchors,
						get = function() return anchorsDict[point[3] or 1] end,
						set = function(info, v) 
							point[3] = anchors[v]; 
							db["pointsDirty"] = true 
							if type(callback) == "function" then
								callback(data)
							end
						end,
						order = 3
					},
					xOfs = {
						name = "X Offset",
						type = "input",
						pattern = "%d",
						get = function() return tostring(point[4] or 0) end,
						set = function(info, v) 
							point[4] = tonumber(v); 
							db["pointsDirty"] = true 
							if type(callback) == "function" then
								callback(data)
							end
						end,
						order = 4
					},
					yOfs = {
						name = "Y Offset",
						type = "input",
						pattern = "%d",
						get = function() return tostring(point[5] or 0) end,
						set = function(info, v) 
							point[5] = tonumber(v);
							db["pointsDirty"] = true 
							if type(callback) == "function" then
								callback(data)
							end
						end,
						order = 4						
					},
					delete = {
						name = "Delete",
						type = "execute",
						func = function()
							tremove(db.points, i)
							if type(callback) == "function" then
								callback(data)
							end
						end,
						order = 5
					}					
				},
				order = i
			}
	end
	return options
end
-- Create a new color
-- @name StarLibWidget.WidgetColor
-- @usage widget:WidgetColor(section, key, C)
-- @param section The configuration section
-- @param key The key within the section -- basic table look up
-- @param C A fresh table where we'll insert color information
-- @return 0 if the color creation fails for some reason, 1 if everything's ok
function LibWidget:WidgetColor(section, key, C)

    C.R = 0;
    C.G = 0;
    C.B = 0;
    C.A = 0;

    local val = self.visitor.CFG:FetchRaw(section, key);

    if (val == nil) then
        return 0;
	end

	if utils:color2RGBA(val, C) <  0 then
		self.error:Print(("widget '%s': ignoring illegal %s color '%s'"):format(self.name, key, val))
		return 0
	end
	
    return 1;
end

